cimport cython
cimport numpy as np


cdef fused real:
    np.float64_t
    np.float32_t

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _index_particles(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] dx,
    np.ndarray[np.uint16_t, ndim=2] out,
):
    r"""
    compute the host cell index (*out*)
    The result is a mapping from particle index to containing cell index.
    Note that cell edges are to be treated as ordained from left to right
    and are padded with ghost layers (one for each side and direction).
    Particles on the other hand are not assumed to be sorted, but they *are*
    assumed to be all contained in the active domain.

    dx is a 3 element array that contains the constant step used in each direction,
    if any, or -1 as a filler value otherwise.

    In directions where a constant step is used, the result is immediate (O(1))
    Otherwise we use a bisection search (O(log(N)))
    """
    cdef Py_ssize_t ipart, particle_count
    cdef np.uint16_t iL, iR, idx
    cdef real x

    cdef int ndim
    if cell_edges_x3.shape[0] > 1:
        ndim = 3
    elif cell_edges_x2.shape[0] > 1:
        ndim = 2
    else:
        ndim = 1

    particle_count = particles_x1.shape[0]

    for ipart in range(particle_count):
        x = particles_x1[ipart]
        if dx[0] > 0:
            out[ipart, 0] = int((x - cell_edges_x1[0]) // dx[0])
        else:
            iL = 1
            iR = cell_edges_x1.shape[0] - 2
            idx = (iL + iR) // 2
            while idx != iL:
                if cell_edges_x1[idx] > x:
                    iR = idx
                else:
                    iL = idx
                idx = (iL + iR) // 2
            out[ipart, 0] = idx

        if ndim < 2:
            continue

        x = particles_x2[ipart]
        if dx[1] > 0:
            out[ipart, 1] = int((x - cell_edges_x2[0]) // dx[1])
        else:
            iL = 1
            iR = cell_edges_x2.shape[0] - 2
            idx = (iL + iR) // 2
            while idx != iL:
                if cell_edges_x2[idx] > x:
                    iR = idx
                else:
                    iL = idx
                idx = (iL + iR) // 2
            out[ipart, 1] = idx

        if ndim < 3:
            continue

        x = particles_x3[ipart]
        if dx[2] > 0:
            out[ipart, 2] = int((x - cell_edges_x3[0]) // dx[2])
        else:
            iL = 1
            iR = cell_edges_x3.shape[0] - 2
            idx = (iL + iR) // 2
            while idx != iL:
                if cell_edges_x3[idx] > x:
                    iR = idx
                else:
                    iL = idx
                idx = (iL + iR) // 2
            out[ipart, 2] = idx



import numpy as np


def _deposit_pic(
    cell_edges_x1,
    cell_edges_x2,
    cell_edges_x3,
    particles_x1,
    particles_x2,
    particles_x3,
    field,
    hci,
    out,
):
    for ipart in range(len(hci)):
        md_idx = tuple(hci[ipart])
        out[md_idx] += field[ipart]


def _deposit_cic(
    cell_edges_x1,
    cell_edges_x2,
    cell_edges_x3,
    particles_x1,
    particles_x2,
    particles_x3,
    field,
    hci,
    out,
):
    return NotImplementedError("Cloud-in-cell deposition method is not implemented yet")


def _deposit_tsc_1D(
    cell_edges_x1,
    cell_edges_x2,
    cell_edges_x3,
    particles_x1,
    particles_x2,
    particles_x3,
    field,
    hci,
    out,
):
    nparticles = hci.shape[0]

    # weight array
    w = np.zeros(3, dtype=field.dtype)

    for ipart in range(nparticles):
        x = particles_x1[ipart]

        ci = hci[ipart, 0]
        d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
        w[0] = 0.5 * (1 - d) ** 2
        w[1] = 0.75 - (d - 0.5) ** 2
        w[2] = 0.5 * d**2

        for k, j in enumerate((ci - 1, ci, ci + 1)):
            out[j] += w[k] * field[ipart]


def _deposit_tsc_2D(
    cell_edges_x1,
    cell_edges_x2,
    cell_edges_x3,
    particles_x1,
    particles_x2,
    particles_x3,
    field,
    hci,
    out,
):
    nparticles = hci.shape[0]

    # weight arrays
    w1 = np.zeros(3, dtype=field.dtype)
    w2 = np.zeros(3, dtype=field.dtype)

    w = np.zeros((3, 3), dtype=field.dtype)

    for ipart in range(nparticles):
        x = particles_x1[ipart]
        ci = hci[ipart, 0]
        d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
        assert d > 0
        w1[0] = 0.5 * (1 - d) ** 2
        w1[1] = 0.75 - (d - 0.5) ** 2
        w1[2] = 0.5 * d**2

        x = particles_x2[ipart]
        cj = hci[ipart, 1]
        d = (x - cell_edges_x2[cj]) / (cell_edges_x2[cj + 1] - cell_edges_x2[cj])
        assert d > 0
        w2[0] = 0.5 * (1 - d) ** 2
        w2[1] = 0.75 - (d - 0.5) ** 2
        w2[2] = 0.5 * d**2

        for i in range(3):
            for j in range(3):
                w[i, j] = w1[i] * w2[j]

        for i, oci in enumerate((ci - 1, ci, ci + 1)):
            for j, ocj in enumerate((cj - 1, cj, cj + 1)):
                out[oci, ocj] += w[i, j] * field[ipart]


def _deposit_tsc_3D(
    cell_edges_x1,
    cell_edges_x2,
    cell_edges_x3,
    particles_x1,
    particles_x2,
    particles_x3,
    field,
    hci,
    out,
):
    nparticles = hci.shape[0]

    # weight arrays
    w1 = np.zeros(3, dtype=field.dtype)
    w2 = np.zeros(3, dtype=field.dtype)
    w3 = np.zeros(3, dtype=field.dtype)

    w = np.zeros((3, 3, 3), dtype=field.dtype)

    for ipart in range(nparticles):
        x = particles_x1[ipart]
        ci = hci[ipart, 0]
        d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
        assert d > 0
        w1[0] = 0.5 * (1 - d) ** 2
        w1[1] = 0.75 - (d - 0.5) ** 2
        w1[2] = 0.5 * d**2

        x = particles_x2[ipart]
        cj = hci[ipart, 1]
        d = (x - cell_edges_x2[cj]) / (cell_edges_x2[cj + 1] - cell_edges_x2[cj])
        assert d > 0
        w2[0] = 0.5 * (1 - d) ** 2
        w2[1] = 0.75 - (d - 0.5) ** 2
        w2[2] = 0.5 * d**2

        x = particles_x3[ipart]
        ck = hci[ipart, 2]
        d = (x - cell_edges_x3[ck]) / (cell_edges_x3[ck + 1] - cell_edges_x3[ck])
        assert d > 0
        w3[0] = 0.5 * (1 - d) ** 2
        w3[1] = 0.75 - (d - 0.5) ** 2
        w3[2] = 0.5 * d**2

        for i in range(3):
            for j in range(3):
                for k in range(3):
                    w[i, j, k] = w1[i] * w2[j] * w3[k]

        for i, oci in enumerate((ci - 1, ci, ci + 1)):
            for j, ocj in enumerate((cj - 1, cj, cj + 1)):
                for k, ock in enumerate((ck - 1, ck, ck + 1)):
                    out[oci, ocj, ock] += w[i, j, k] * field[ipart]



def _index_particles(
    cell_edges_x1,
    cell_edges_x2,
    cell_edges_x3,
    particles_x1,
    particles_x2,
    particles_x3,
    dx,
    out,
):
    if cell_edges_x3.shape[0] > 1:
        ndim = 3
    elif cell_edges_x2.shape[0] > 1:
        ndim = 2
    else:
        ndim = 1

    particle_count = particles_x1.shape[0]

    for ipart in range(particle_count):
        x = particles_x1[ipart]
        if dx[0] > 0:
            out[ipart, 0] = int((x - cell_edges_x1[0]) // dx[0])
        else:
            iL = 1
            iR = cell_edges_x1.shape[0] - 2
            idx = (iL + iR) // 2
            while idx != iL:
                if cell_edges_x1[idx] > x:
                    iR = idx
                else:
                    iL = idx
                idx = (iL + iR) // 2
            out[ipart, 0] = idx

        if ndim < 2:
            continue

        x = particles_x2[ipart]
        if dx[1] > 0:
            out[ipart, 1] = int((x - cell_edges_x2[0]) // dx[1])
        else:
            iL = 1
            iR = cell_edges_x2.shape[00] - 2
            idx = (iL + iR) // 2
            while idx != iL:
                if cell_edges_x2[idx] > x:
                    iR = idx
                else:
                    iL = idx
                idx = (iL + iR) // 2
            out[ipart, 1] = idx

        if ndim < 3:
            continue

        x = particles_x3[ipart]
        if dx[2] > 0:
            out[ipart, 2] = int((x - cell_edges_x3[0]) // dx[2])
        else:
            iL = 1
            iR = cell_edges_x3.shape[0] - 2
            idx = (iL + iR) // 2
            while idx != iL:
                if cell_edges_x3[idx] > x:
                    iR = idx
                else:
                    iL = idx
                idx = (iL + iR) // 2
            out[ipart, 2] = idx



from __future__ import annotations

from .types import Dataset
from .types import FieldMap
from .types import Geometry
from .types import Grid
from .types import ParticleSet

_geometry_names: dict[str, Geometry] = {g.name.lower(): g for g in Geometry}


def load(
    *,
    geometry: str = "cartesian",
    grid: dict[str, FieldMap] | None = None,
    particles: dict[str, FieldMap] | None = None,
) -> Dataset:

    if geometry in _geometry_names:
        _geometry = _geometry_names[geometry]
    else:
        raise ValueError(
            f"unknown geometry {geometry!r}, expected any of {tuple(_geometry_names.keys())}"
        )

    _grid: Grid | None = None
    if grid is not None:
        if "cell_edges" not in grid:
            raise ValueError("grid dictionary missing required key 'cell_edges'")
        _grid = Grid(
            _geometry, cell_edges=grid["cell_edges"], fields=grid.get("fields", {})
        )

    _particles: ParticleSet | None = None
    if particles is not None:
        if "coordinates" not in particles:
            raise ValueError("particles dictionary missing required key 'coordinates'")
        _particles = ParticleSet(
            _geometry,
            coordinates=particles["coordinates"],
            fields=particles.get("fields", {}),
        )

    return Dataset(geometry=_geometry, grid=_grid, particles=_particles)



cimport cython
cimport numpy as np


cdef fused real:
    np.float64_t
    np.float32_t

@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_pic_1D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=1] out,
):
    cdef Py_ssize_t ipart, particle_count
    particle_count = hci.shape[0]

    cdef int i

    cdef np.uint16_t[:, :] hci_v = hci
    cdef real[:] field_v = field
    cdef real[:] out_v = out

    for ipart in range(particle_count):
        i = hci_v[ipart][0]
        out_v[i] += field_v[ipart]

@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_pic_2D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=2] out,
):
    cdef Py_ssize_t ipart, particle_count
    particle_count = hci.shape[0]

    cdef int i, j

    cdef np.uint16_t[:, :] hci_v = hci
    cdef real[:] field_v = field
    cdef real[:, :] out_v = out

    for ipart in range(particle_count):
        i = hci_v[ipart][0]
        j = hci_v[ipart][1]
        out_v[i][j] += field_v[ipart]


@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_pic_3D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=3] out,
):
    cdef Py_ssize_t ipart, particle_count
    particle_count = hci.shape[0]

    cdef int i, j, k

    cdef np.uint16_t[:, :] hci_v = hci
    cdef real[:] field_v = field
    cdef real[:, :, :] out_v = out

    for ipart in range(particle_count):
        i = hci_v[ipart][0]
        j = hci_v[ipart][1]
        k = hci_v[ipart][2]
        out_v[i][j][k] += field_v[ipart]


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_tsc_1D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=1] out,
):
    cdef Py_ssize_t ipart, particle_count, i, oci
    particle_count = hci.shape[0]

    cdef real x, d
    cdef np.uint16_t ci

    cdef real[:] field_v = field
    cdef real[:] out_v = out

    # weight arrays
    cdef real[3] w

    for ipart in range(particle_count):
        x = particles_x1[ipart]
        ci = hci[ipart, 0]
        d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
        w[0] = 0.5 * (1 - d) ** 2
        w[1] = 0.75 - (d - 0.5) ** 2
        w[2] = 0.5 * d**2

        for i in range(3):
            oci = ci - 1 + i
            out_v[oci] += w[i] * field_v[ipart]

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_tsc_2D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=2] out,
):
    cdef Py_ssize_t ipart, particle_count, i, j, oci, ocj
    particle_count = hci.shape[0]

    cdef real x, d
    cdef np.uint16_t ci, cj

    cdef real[:] field_v = field
    cdef real[:, :] out_v = out

    # weight arrays
    cdef real[3] w1, w2
    cdef real[3][3] w

    for ipart in range(particle_count):
        x = particles_x1[ipart]
        ci = hci[ipart, 0]
        d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
        w1[0] = 0.5 * (1 - d) ** 2
        w1[1] = 0.75 - (d - 0.5) ** 2
        w1[2] = 0.5 * d**2

        x = particles_x2[ipart]
        cj = hci[ipart, 1]
        d = (x - cell_edges_x2[cj]) / (cell_edges_x2[cj + 1] - cell_edges_x2[cj])
        w2[0] = 0.5 * (1 - d) ** 2
        w2[1] = 0.75 - (d - 0.5) ** 2
        w2[2] = 0.5 * d**2

        for i in range(3):
            for j in range(3):
                w[i][j] = w1[i] * w2[j]

        for i in range(3):
            oci = ci - 1 + i
            for j in range(3):
                ocj = cj - 1 + j
                out_v[oci, ocj] += w[i][j] * field_v[ipart]


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _deposit_tsc_3D(
    np.ndarray[real, ndim=1] cell_edges_x1,
    np.ndarray[real, ndim=1] cell_edges_x2,
    np.ndarray[real, ndim=1] cell_edges_x3,
    np.ndarray[real, ndim=1] particles_x1,
    np.ndarray[real, ndim=1] particles_x2,
    np.ndarray[real, ndim=1] particles_x3,
    np.ndarray[real, ndim=1] field,
    np.ndarray[np.uint16_t, ndim=2] hci,
    np.ndarray[real, ndim=3] out,
):
    cdef Py_ssize_t ipart, particle_count, i, j, k, oci, ocj, ock
    particle_count = hci.shape[0]

    cdef real x, d
    cdef np.uint16_t ci, cj, ck

    cdef real[:] field_v = field
    cdef real[:, :, :] out_v = out

    # weight arrays
    cdef real[3] w1, w2, w3
    cdef real[3][3][3] w

    for ipart in range(particle_count):
        x = particles_x1[ipart]
        ci = hci[ipart, 0]
        d = (x - cell_edges_x1[ci]) / (cell_edges_x1[ci + 1] - cell_edges_x1[ci])
        w1[0] = 0.5 * (1 - d) ** 2
        w1[1] = 0.75 - (d - 0.5) ** 2
        w1[2] = 0.5 * d**2

        x = particles_x2[ipart]
        cj = hci[ipart, 1]
        d = (x - cell_edges_x2[cj]) / (cell_edges_x2[cj + 1] - cell_edges_x2[cj])
        w2[0] = 0.5 * (1 - d) ** 2
        w2[1] = 0.75 - (d - 0.5) ** 2
        w2[2] = 0.5 * d**2

        x = particles_x3[ipart]
        ck = hci[ipart, 2]
        d = (x - cell_edges_x3[ck]) / (cell_edges_x3[ck + 1] - cell_edges_x3[ck])
        w3[0] = 0.5 * (1 - d) ** 2
        w3[1] = 0.75 - (d - 0.5) ** 2
        w3[2] = 0.5 * d**2

        for i in range(3):
            for j in range(3):
                for k in range(3):
                    w[i][j][k] = w1[i] * w2[j] * w3[k]

        for i in range(3):
            oci = ci - 1 + i
            for j in range(3):
                ocj = cj - 1 + j
                for k in range(3):
                    ock = ck - 1 + k
                    out_v[oci, ocj, ock] += w[i][j][k] * field_v[ipart]



import numpy
from Cython.Build import cythonize
from distutils.extension import Extension
from setuptools import setup

ext_modules = ["gpgi/clib/_indexing.pyx"]


setup(
    ext_modules=cythonize(
        [
            Extension(
                "gpgi.clib._indexing",
                ["gpgi/clib/_indexing.pyx"],
                include_dirs=[numpy.get_include()],
            ),
            Extension(
                "gpgi.clib._deposition_methods",
                ["gpgi/clib/_deposition_methods.pyx"],
                include_dirs=[numpy.get_include()],
            ),
        ],
        compiler_directives={"language_level": 3},
        annotate=True,
    ),
)



from __future__ import annotations

import enum
from abc import ABC
from abc import abstractmethod
from functools import cached_property
from functools import reduce
from itertools import chain
from time import monotonic_ns
from typing import Any
from typing import Protocol

import numpy as np


class Geometry(enum.Enum):
    CARTESIAN = enum.auto()
    POLAR = enum.auto()
    CYLINDRICAL = enum.auto()
    SPHERICAL = enum.auto()


class DepositionMethod(enum.Enum):
    PARTICLE_IN_CELL = enum.auto()
    CLOUD_IN_CELL = enum.auto()
    TRIANGULAR_SHAPED_CLOUD = enum.auto()


_deposition_method_names: dict[str, DepositionMethod] = {
    m.name.lower(): m for m in DepositionMethod
} | {"".join([w[0] for w in m.name.split("_")]).lower(): m for m in DepositionMethod}


Name = str
FieldMap = dict[Name, np.ndarray]


class GeometricData(Protocol):
    geometry: Geometry
    axes: tuple[Name, ...]


class CoordinateData(Protocol):
    geometry: Geometry
    axes: tuple[Name, ...]
    coordinates: FieldMap


class ValidatorMixin(GeometricData, ABC):
    def __init__(self) -> None:
        self._validate()

    @abstractmethod
    def _validate(self) -> None:
        ...

    def _validate_fieldmaps(
        self,
        *fmaps: FieldMap | None,
        require_shape_equality: bool = False,
        require_sorted: bool = False,
        **required_attrs: Any,
    ) -> None:
        _reference_shape: tuple[int, ...] | None = None
        _reference_field_name: str
        for fmap in fmaps:
            if fmap is None:
                continue  # pragma: no cover
            for name, data in fmap.items():
                if require_shape_equality:
                    if _reference_shape is None:
                        _reference_shape = data.shape
                        _reference_field_name = name
                    elif data.shape != _reference_shape:
                        raise ValueError(
                            f"Fields {name!r} and {_reference_field_name!r} "
                            f"have mismatching shapes {data.shape} and {_reference_shape}"
                        )

                if require_sorted:
                    _a = data[0]
                    for i, _b in enumerate(data[1:], start=1):
                        if _a > _b:
                            raise ValueError(
                                f"Field {name!r} is not properly sorted by ascending order. "
                                f"Got {_a} (index {i-1}) > {_b} (index {i})"
                            )
                        _a = _b

                if not required_attrs:
                    continue  # pragma: no cover
                for attr, expected in required_attrs.items():
                    if (actual := getattr(data, attr)) != expected:
                        raise ValueError(
                            f"Field {name!r} has incorrect {attr} {actual} "
                            f"(expected {expected})"
                        )

    def _validate_geometry(self) -> None:
        known_axes: dict[Geometry, tuple[Name, Name, Name]] = {
            Geometry.CARTESIAN: ("x", "y", "z"),
            Geometry.POLAR: ("radius", "z", "azimuth"),
            Geometry.CYLINDRICAL: ("radius", "azimuth", "z"),
            Geometry.SPHERICAL: ("radius", "colatitude", "azimuth"),
        }
        if self.geometry not in known_axes:
            # TODO: when Python 3.10 is required, refactor as a match/case block
            # and check that the default case is unreacheable at type check time
            raise ValueError(
                f"Unknown geometry {self.geometry.name.lower()!r}"
            )  # pragma: no cover

        axes = known_axes[self.geometry]
        for expected, actual in zip(axes, self.axes):
            if actual != expected:
                raise ValueError(
                    f"Got invalid axis {actual!r} with geometry {self.geometry.name.lower()!r}"
                )


class CoordinateValidatorMixin(ValidatorMixin, CoordinateData, ABC):
    def _validate_coordinates(self) -> None:
        known_limits: dict[Name, tuple[float | None, float | None]] = {
            "radius": (0, None),
            "azimuth": (0, 2 * np.pi),
            "colatitude": (0, np.pi),
        }
        for axis, coord in self.coordinates.items():
            lims = known_limits.get(axis)
            if lims is None:
                continue
            xmin, xmax = lims
            if xmin is not None and (cmin := np.min(coord)) < xmin:
                raise ValueError(
                    f"Invalid coordinate data for axis {axis!r} {cmin} "
                    f"(minimal allowed value is {xmin})"
                )
            if xmax is not None and (cmax := np.max(coord)) > xmax:
                raise ValueError(
                    f"Invalid coordinate data for axis {axis!r} {cmax} "
                    f"(maximal allowed value is {xmax})"
                )


class Grid(CoordinateValidatorMixin):
    def __init__(
        self,
        geometry: Geometry,
        cell_edges: FieldMap,
        fields: FieldMap,
    ) -> None:
        self.geometry = geometry
        self.coordinates = cell_edges
        self.fields = fields

        self.axes = tuple(self.coordinates.keys())
        super().__init__()

        self._dx = np.full((3,), -1, dtype=self.coordinates[self.axes[0]].dtype)
        for i, ax in enumerate(self.axes):
            if np.diff(self.coordinates[ax]).std() < 1e-16:
                # got a constant step in this direction, store it
                self._dx[i] = self.coordinates[ax][1] - self.coordinates[ax][0]

    def _validate(self) -> None:
        self._validate_geometry()
        self._validate_coordinates()
        self._validate_fieldmaps(self.cell_edges, ndim=1, require_sorted=True)
        self._validate_fieldmaps(
            self.fields,
            size=self.size,
            ndim=self.ndim,
            shape=self.shape,
        )

    @property
    def cell_edges(self) -> FieldMap:
        return self.coordinates

    @cached_property
    def cell_centers(self) -> FieldMap:
        return {ax: 0.5 * (arr[1:] + arr[:-1]) for ax, arr in self.coordinates.items()}

    @cached_property
    def cell_widths(self) -> FieldMap:
        return {ax: np.diff(arr) for ax, arr in self.coordinates.items()}

    @cached_property
    def shape(self) -> tuple[int, ...]:
        return tuple(len(_) - 1 for _ in self.cell_edges.values())

    @cached_property
    def _padded_shape(self) -> tuple[int, ...]:
        return tuple(len(_) + 1 for _ in self.cell_edges.values())

    @property
    def size(self) -> int:
        return reduce(int.__mul__, self.shape)

    @property
    def ndim(self) -> int:
        return len(self.axes)


class ParticleSet(CoordinateValidatorMixin):
    def __init__(
        self,
        geometry: Geometry,
        coordinates: FieldMap,
        fields: FieldMap,
    ) -> None:
        self.geometry = geometry
        self.coordinates = coordinates
        self.fields = fields

        self.axes = tuple(self.coordinates.keys())
        super().__init__()

    def _validate(self) -> None:
        self._validate_geometry()
        self._validate_coordinates()
        self._validate_fieldmaps(
            self.coordinates, self.fields, require_shape_equality=True, ndim=1
        )

    @property
    def count(self) -> int:
        return len(next(iter(self.coordinates.values())))

    @property
    def ndim(self) -> int:
        return len(self.axes)


class Dataset(ValidatorMixin):
    def __init__(
        self,
        *,
        geometry: Geometry = Geometry.CARTESIAN,
        grid: Grid | None = None,
        particles: ParticleSet | None = None,
    ) -> None:
        self.geometry = geometry
        self.grid = grid
        self.particles = particles

        if self.grid is not None:
            self.axes = self.grid.axes
        elif self.particles is not None:
            self.axes = self.particles.axes
        else:
            raise TypeError(
                "Cannot instantiate empty dataset. "
                "Grid and/or particle data must be provided"
            )
        super().__init__()

    def _validate(self) -> None:
        if self.grid is None or self.particles is None:
            return
        for ax, edges in self.grid.cell_edges.items():
            domain_left = edges[0]
            domain_right = edges[-1]
            for x in self.particles.coordinates[ax]:
                if x < domain_left:
                    raise ValueError(f"Got particle at {ax}={x} < {domain_left=}")
                if x > domain_right:
                    raise ValueError(f"Got particle at {ax}={x} > {domain_right=}")

        unique_dts = sorted(
            {
                arr.dtype
                for arr in chain(
                    self.grid.coordinates.values(),
                    self.grid.fields.values(),
                    self.particles.coordinates.values(),
                    self.particles.fields.values(),
                )
            }
        )
        if len(unique_dts) > 1:
            raise TypeError(f"Got mixed data types ({unique_dts})")

    def _get_padded_cell_edges(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        if self.grid is None:
            raise RuntimeError(
                "Something took a wrong turn, please report this."
            )  # pragma: no cover
        edges = iter(self.grid.cell_edges.values())

        def pad(a: np.ndarray) -> np.ndarray:
            dx = a[1] - a[0]
            return np.concatenate([[a[0] - dx], a, [a[-1] + dx]])

        x1 = next(edges)
        cell_edges_x1 = pad(x1)
        DTYPE = cell_edges_x1.dtype

        cell_edges_x2 = cell_edges_x3 = np.empty(0, DTYPE)
        if self.grid.ndim >= 2:
            cell_edges_x2 = pad(next(edges))
        if self.grid.ndim == 3:
            cell_edges_x3 = pad(next(edges))

        return cell_edges_x1, cell_edges_x2, cell_edges_x3

    def _get_3D_particle_coordinates(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        if self.grid is None or self.particles is None:
            raise RuntimeError(
                "Something took a wrong turn, please report this."
            )  # pragma: no cover

        particle_coords = iter(self.particles.coordinates.values())
        particles_x1 = next(particle_coords)
        DTYPE = particles_x1.dtype

        particles_x2 = particles_x3 = np.empty(0, DTYPE)
        if self.grid.ndim >= 2:
            particles_x2 = next(particle_coords)
        if self.grid.ndim == 3:
            particles_x3 = next(particle_coords)

        return particles_x1, particles_x2, particles_x3

    def _setup_host_cell_index(self, verbose: bool = False) -> None:
        if hasattr(self, "_hci") or self.particles is None or self.grid is None:
            # this line is hard to cover by testing just public api
            # because it's a pure performance optimization with no other observable effect
            return  # pragma: no cover
        from .clib._indexing import _index_particles  # type: ignore [import]

        self._hci = np.empty((self.particles.count, self.grid.ndim), dtype="uint16")

        tstart = monotonic_ns()
        _index_particles(
            *self._get_padded_cell_edges(),
            *self._get_3D_particle_coordinates(),
            dx=self.grid._dx,
            out=self._hci,
        )
        tstop = monotonic_ns()
        if verbose:
            print(
                f"Indexed {self.particles.count:.4g} particles in {(tstop-tstart)/1e9:.2f} s"
            )

    def deposit(
        self, particle_field_key: Name, /, *, method: Name, verbose: bool = False
    ) -> np.ndarray:
        from .clib._deposition_methods import _deposit_pic_1D  # type: ignore [import]
        from .clib._deposition_methods import _deposit_pic_2D  # type: ignore [import]
        from .clib._deposition_methods import _deposit_pic_3D  # type: ignore [import]
        from .clib._deposition_methods import _deposit_tsc_1D  # type: ignore [import]
        from .clib._deposition_methods import _deposit_tsc_2D  # type: ignore [import]
        from .clib._deposition_methods import _deposit_tsc_3D  # type: ignore [import]

        if not hasattr(self, "_cache"):
            self._cache: dict[tuple[Name, DepositionMethod], np.ndarray] = {}

        if method in _deposition_method_names:
            mkey = _deposition_method_names[method]
        else:
            raise ValueError(
                f"Unknown deposition method {method!r}, "
                f"expected any of {tuple(_deposition_method_names.keys())}"
            )

        if (particle_field_key, mkey) in self._cache:
            return self._cache[particle_field_key, mkey]

        if self.grid is None:
            raise TypeError("Cannot deposit particle fields on a grid-less dataset")
        if self.particles is None:
            raise TypeError("Cannot deposit particle fields on a particle-less dataset")
        if not self.particles.fields:
            raise TypeError("There are no particle fields")
        if particle_field_key not in self.particles.fields:
            raise ValueError(f"Unknown particle field {particle_field_key!r}")

        # deactivating type checking for deposition methods because
        # they may be ported to Cyhton later
        known_methods: dict[DepositionMethod, list[Any]] = {
            DepositionMethod.PARTICLE_IN_CELL: [
                _deposit_pic_1D,
                _deposit_pic_2D,
                _deposit_pic_3D,
            ],
            DepositionMethod.TRIANGULAR_SHAPED_CLOUD: [
                _deposit_tsc_1D,
                _deposit_tsc_2D,
                _deposit_tsc_3D,
            ],
        }
        if mkey not in known_methods:
            raise NotImplementedError(f"method {method} is not implemented yet")

        field = np.array(self.particles.fields[particle_field_key])
        padded_ret_array = np.zeros(self.grid._padded_shape, dtype=field.dtype)
        self._setup_host_cell_index(verbose=verbose)

        func = known_methods[mkey][self.grid.ndim - 1]
        tstart = monotonic_ns()
        func(
            *self._get_padded_cell_edges(),
            *self._get_3D_particle_coordinates(),
            field,
            self._hci,
            padded_ret_array,
        )
        tstop = monotonic_ns()
        if verbose:
            print(
                f"Deposited {self.particles.count:.4g} particles in {(tstop-tstart)/1e9:.2f} s"
            )

        # boundary conditions treatment should be performed here
        # ...

        # remove ghost layer padding
        if self.grid.ndim == 1:
            ret_array = padded_ret_array[1:-1]
        elif self.grid.ndim == 2:
            ret_array = padded_ret_array[1:-1, 1:-1]
        elif self.grid.ndim == 3:
            ret_array = padded_ret_array[1:-1, 1:-1, 1:-1]
        self._cache[particle_field_key, mkey] = ret_array
        return ret_array
